@@ -17,20 +17,23 @@ import sys
 import threading
 import time
 import warnings
-
-from types import FrameType
-from typing import cast, Any, Callable, IO
 from collections.abc import Iterable, Iterator
+from types import FrameType
+from typing import IO, Any, Callable, Union, cast
 
 from coverage import env
 from coverage.annotate import AnnotateReporter
 from coverage.collector import Collector
-from coverage.config import CoverageConfig, read_coverage_config
-from coverage.context import should_start_context_test_function, combine_context_switchers
-from coverage.core import Core, HAS_CTRACER
+from coverage.config import CoverageConfig, deserialize_config, read_coverage_config
+from coverage.context import combine_context_switchers, should_start_context_test_function
+from coverage.core import CTRACER_FILE, Core
 from coverage.data import CoverageData, combine_parallel_data
 from coverage.debug import (
-    DebugControl, NoDebugging, short_stack, write_formatted_info, relevant_environment_display,
+    DebugControl,
+    NoDebugging,
+    relevant_environment_display,
+    short_stack,
+    write_formatted_info,
 )
 from coverage.disposition import disposition_debug_msg
 from coverage.exceptions import ConfigError, CoverageException, CoverageWarning, PluginError
@@ -39,18 +42,30 @@ from coverage.html import HtmlReporter
 from coverage.inorout import InOrOut
 from coverage.jsonreport import JsonReporter
 from coverage.lcovreport import LcovReporter
-from coverage.misc import bool_or_none, join_regex
-from coverage.misc import DefaultValue, ensure_dir_for_file, isolate_module
+from coverage.misc import (
+    DefaultValue,
+    bool_or_none,
+    ensure_dir_for_file,
+    isolate_module,
+    join_regex,
+)
 from coverage.multiproc import patch_multiprocessing
+from coverage.patch import apply_patches
 from coverage.plugin import FileReporter
-from coverage.plugin_support import Plugins
+from coverage.plugin_support import Plugins, TCoverageInit
 from coverage.python import PythonFileReporter
 from coverage.report import SummaryReporter
 from coverage.report_core import render_report
 from coverage.results import Analysis, analysis_from_file_reporter
 from coverage.types import (
-    FilePath, TConfigurable, TConfigSectionIn, TConfigValueIn, TConfigValueOut,
-    TFileDisposition, TLineNo, TMorf,
+    FilePath,
+    TConfigSectionIn,
+    TConfigurable,
+    TConfigValueIn,
+    TConfigValueOut,
+    TFileDisposition,
+    TLineNo,
+    TMorf,
 )
 from coverage.xmlreport import XmlReporter
 
@@ -74,6 +89,7 @@ def override_config(cov: Coverage, **kwargs: TConfigValueIn) -> Iterator[None]:
 
 DEFAULT_DATAFILE = DefaultValue("MISSING")
 _DEFAULT_DATAFILE = DEFAULT_DATAFILE  # Just in case, for backwards compatibility
+CONFIG_FROM_ENVIRONMENT = ":envvar:"
 
 class Coverage(TConfigurable):
     """Programmatic access to coverage.py.
@@ -131,6 +147,7 @@ class Coverage(TConfigurable):
         config_file: FilePath | bool = True,
         source: Iterable[str] | None = None,
         source_pkgs: Iterable[str] | None = None,
+        source_dirs: Iterable[str] | None = None,
         omit: str | Iterable[str] | None = None,
         include: str | Iterable[str] | None = None,
         debug: Iterable[str] | None = None,
@@ -138,6 +155,7 @@ class Coverage(TConfigurable):
         check_preimported: bool = False,
         context: str | None = None,
         messages: bool = False,
+        plugins: Iterable[Callable[..., None]] | None = None,
     ) -> None:
         """
         Many of these arguments duplicate and override values that can be
@@ -187,6 +205,10 @@ class Coverage(TConfigurable):
         `source`, but can be used to name packages where the name can also be
         interpreted as a file path.
 
+        `source_dirs` is a list of file paths. It works the same as
+        `source`, but raises an error if the path doesn't exist, rather
+        than being treated as a package name.
+
         `include` and `omit` are lists of file name patterns. Files that match
         `include` will be measured, files that match `omit` will not.  Each
         will also accept a single string argument.
@@ -211,6 +233,11 @@ class Coverage(TConfigurable):
         If `messages` is true, some messages will be printed to stdout
         indicating what is happening.
 
+        If `plugins` are passed, they are an iterable of function objects
+        accepting a `reg` object to register plugins, as described in
+        :ref:`api_plugin`.  When they are provided, they will override the
+        plugins found in the coverage configuration file.
+
         .. versionadded:: 4.0
             The `concurrency` parameter.
 
@@ -226,6 +253,11 @@ class Coverage(TConfigurable):
         .. versionadded:: 6.0
             The `messages` parameter.
 
+        .. versionadded:: 7.7
+            The `plugins` parameter.
+
+        .. versionadded:: 7.8
+            The `source_dirs` parameter.
         """
         # Start self.config as a usable default configuration. It will soon be
         # replaced with the real configuration.
@@ -249,9 +281,12 @@ class Coverage(TConfigurable):
         self._warn_no_data = True
         self._warn_unimported_source = True
         self._warn_preimported_source = check_preimported
-        self._no_warn_slugs: list[str] = []
+        self._no_warn_slugs: set[str] = set()
         self._messages = messages
 
+        # If we're invoked from a .pth file, we shouldn't try to make another one.
+        self._make_pth_file = True
+
         # A record of all the warnings that have been issued.
         self._warnings: list[str] = []
 
@@ -260,7 +295,9 @@ class Coverage(TConfigurable):
         self._debug: DebugControl = NoDebugging()
         self._inorout: InOrOut | None = None
         self._plugins: Plugins = Plugins()
+        self._plugin_override = cast(Union[Iterable[TCoverageInit], None], plugins)
         self._data: CoverageData | None = None
+        self._data_to_close: list[CoverageData] = []
         self._core: Core | None = None
         self._collector: Collector | None = None
         self._metacov = False
@@ -280,26 +317,30 @@ class Coverage(TConfigurable):
         self._should_write_debug = True
 
         # Build our configuration from a number of sources.
-        if not isinstance(config_file, bool):
-            config_file = os.fspath(config_file)
-        self.config = read_coverage_config(
-            config_file=config_file,
-            warn=self._warn,
-            data_file=data_file,
-            cover_pylib=cover_pylib,
-            timid=timid,
-            branch=branch,
-            parallel=bool_or_none(data_suffix),
-            source=source,
-            source_pkgs=source_pkgs,
-            run_omit=omit,
-            run_include=include,
-            debug=debug,
-            report_omit=omit,
-            report_include=include,
-            concurrency=concurrency,
-            context=context,
-        )
+        if config_file == CONFIG_FROM_ENVIRONMENT:
+            self.config = deserialize_config(cast(str, os.getenv("COVERAGE_PROCESS_CONFIG")))
+        else:
+            if not isinstance(config_file, bool):
+                config_file = os.fspath(config_file)
+            self.config = read_coverage_config(
+                config_file=config_file,
+                warn=self._warn,
+                data_file=data_file,
+                cover_pylib=cover_pylib,
+                timid=timid,
+                branch=branch,
+                parallel=bool_or_none(data_suffix),
+                source=source,
+                source_pkgs=source_pkgs,
+                source_dirs=source_dirs,
+                run_omit=omit,
+                run_include=include,
+                debug=debug,
+                report_omit=omit,
+                report_include=include,
+                concurrency=concurrency,
+                context=context,
+            )
 
         # If we have subprocess measurement happening automatically, then we
         # want any explicit creation of a Coverage object to mean, this process
@@ -340,7 +381,11 @@ class Coverage(TConfigurable):
             self._file_mapper = relative_filename
 
         # Load plugins
-        self._plugins = Plugins.load_plugins(self.config.plugins, self.config, self._debug)
+        self._plugins = Plugins(self._debug)
+        if self._plugin_override:
+            self._plugins.load_from_callables(self._plugin_override)
+        else:
+            self._plugins.load_from_config(self.config.plugins, self.config)
 
         # Run configuring plugins.
         for plugin in self._plugins.configurers:
@@ -424,7 +469,7 @@ class Coverage(TConfigurable):
 
         """
         if not self._no_warn_slugs:
-            self._no_warn_slugs = list(self.config.disable_warnings)
+            self._no_warn_slugs = set(self.config.disable_warnings)
 
         if slug in self._no_warn_slugs:
             # Don't issue the warning
@@ -439,7 +484,7 @@ class Coverage(TConfigurable):
 
         if once:
             assert slug is not None
-            self._no_warn_slugs.append(slug)
+            self._no_warn_slugs.add(slug)
 
     def _message(self, msg: str) -> None:
         """Write a message to the user, if configured to do so."""
@@ -534,7 +579,8 @@ class Coverage(TConfigurable):
 
         self._core = Core(
             warn=self._warn,
-            timid=self.config.timid,
+            config=self.config,
+            dynamic_contexts=(should_start_context is not None),
             metacov=self._metacov,
         )
         self._collector = Collector(
@@ -619,6 +665,7 @@ class Coverage(TConfigurable):
                 debug=self._debug,
                 no_disk=self._no_disk,
             )
+            self._data_to_close.append(self._data)
 
     def start(self) -> None:
         """Start measuring code coverage.
@@ -654,6 +701,8 @@ class Coverage(TConfigurable):
         if self._auto_load:
             self.load()
 
+        apply_patches(self, self.config, self._debug, make_pth_file=self._make_pth_file)
+
         self._collector.start()
         self._started = True
         self._instances.append(self)
@@ -689,6 +738,8 @@ class Coverage(TConfigurable):
             self.stop()
         if self._auto_save or event == "sigterm":
             self.save()
+        for d in self._data_to_close:
+            d.close(force=True)
 
     def _on_sigterm(self, signum_unused: int, frame_unused: FrameType | None) -> None:
         """A handler for signal.SIGTERM."""
@@ -935,6 +986,7 @@ class Coverage(TConfigurable):
             analysis.missing_formatted(),
         )
 
+    @functools.lru_cache(maxsize=1)
     def _analyze(self, morf: TMorf) -> Analysis:
         """Analyze a module or file.  Private for now."""
         self._init()
@@ -945,6 +997,20 @@ class Coverage(TConfigurable):
         filename = self._file_mapper(file_reporter.filename)
         return analysis_from_file_reporter(data, self.config.precision, file_reporter, filename)
 
+    def branch_stats(self, morf: TMorf) -> dict[TLineNo, tuple[int, int]]:
+        """Get branch statistics about a module.
+
+        `morf` is a module or a file name.
+
+        Returns a dict mapping line numbers to a tuple:
+        (total_exits, taken_exits).
+
+        .. versionadded:: 7.7
+
+        """
+        analysis = self._analyze(morf)
+        return analysis.branch_stats()
+
     @functools.lru_cache(maxsize=1)
     def _get_file_reporter(self, morf: TMorf) -> FileReporter:
         """Get a FileReporter for a module or file name."""
@@ -1004,6 +1070,7 @@ class Coverage(TConfigurable):
             if self._data is not None:
                 mapped_data.update(self._data, map_path=self._make_aliases().map)
             self._data = mapped_data
+            self._data_to_close.append(mapped_data)
 
     def report(
         self,
@@ -1298,7 +1365,7 @@ class Coverage(TConfigurable):
             ("coverage_version", covmod.__version__),
             ("coverage_module", covmod.__file__),
             ("core", self._collector.tracer_name() if self._collector is not None else "-none-"),
-            ("CTracer", "available" if HAS_CTRACER else "unavailable"),
+            ("CTracer", f"available from {CTRACER_FILE}" if CTRACER_FILE else "unavailable"),
             ("plugins.file_tracers", plugin_info(self._plugins.file_tracers)),
             ("plugins.configurers", plugin_info(self._plugins.configurers)),
             ("plugins.context_switchers", plugin_info(self._plugins.context_switchers)),
@@ -1312,7 +1379,8 @@ class Coverage(TConfigurable):
             ("python", sys.version.replace("\n", "")),
             ("platform", platform.platform()),
             ("implementation", platform.python_implementation()),
-            ("gil_enabled", getattr(sys, '_is_gil_enabled', lambda: True)()),
+            ("build", platform.python_build()),
+            ("gil_enabled", getattr(sys, "_is_gil_enabled", lambda: True)()),
             ("executable", sys.executable),
             ("def_encoding", sys.getdefaultencoding()),
             ("fs_encoding", sys.getfilesystemencoding()),
@@ -1349,24 +1417,19 @@ def process_startup() -> Coverage | None:
     measurement is started.  The value of the variable is the config file
     to use.
 
-    There are two ways to configure your Python installation to invoke this
-    function when Python starts:
-
-    #. Create or append to sitecustomize.py to add these lines::
-
-        import coverage
-        coverage.process_startup()
-
-    #. Create a .pth file in your Python installation containing::
-
-        import coverage; coverage.process_startup()
+    For details, see https://coverage.readthedocs.io/en/latest/subprocess.html.
 
     Returns the :class:`Coverage` instance that was started, or None if it was
     not started by this call.
 
     """
     cps = os.getenv("COVERAGE_PROCESS_START")
-    if not cps:
+    config_data = os.getenv("COVERAGE_PROCESS_CONFIG")
+    if cps is not None:
+        config_file = cps
+    elif config_data is not None:
+        config_file = CONFIG_FROM_ENVIRONMENT
+    else:
         # No request for coverage, nothing to do.
         return None
 
@@ -1381,15 +1444,16 @@ def process_startup() -> Coverage | None:
 
     if hasattr(process_startup, "coverage"):
         # We've annotated this function before, so we must have already
-        # started coverage.py in this process.  Nothing to do.
+        # auto-started coverage.py in this process.  Nothing to do.
         return None
 
-    cov = Coverage(config_file=cps)
+    cov = Coverage(config_file=config_file)
     process_startup.coverage = cov      # type: ignore[attr-defined]
     cov._warn_no_data = False
     cov._warn_unimported_source = False
     cov._warn_preimported_source = False
     cov._auto_save = True
+    cov._make_pth_file = False
     cov.start()
 
     return cov
